#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <new> // for placement new

#include "assertion.h"
#include "ilist.h"
#include "timer.h"

template <class T>
T max(T a, T b) { return a > b ? a : b; }

template <class T>
inline bool maximize(T& a, T b) {
  return (b > a) ? (a = b, true) : false;
}

template <class T>
inline bool minimize(T& a, T b) {
  return (b < a) ? (a = b, true) : false;
}

class Word;
class Break;
typedef ilist<Word,slnk,false>  WordList;
typedef ilist<Break,slnk,false> BreakList;

class Word : public slnk {
  char m_word[30]; // word or syllable FIXME: 30
  int  m_size;     // # of chars
  int  m_width;    // sum of char-widths
  int  m_accu;     // accumulated length for sequence of words before
  bool m_syllable; // result of hyphenation ?
public:
  inline Word(const char* w, int* char_width, bool s)
         : m_size(strlen(w)), 
           m_width(0),
           m_accu(0),
           m_syllable(s) {
    assertion(m_size + 1 <= int(sizeof(m_word)),"word too long");
    strcpy(m_word,w);
    if (char_width) set_width(char_width);
    // printf("%s %d %d\n",m_word,m_size,m_width);
  }
  inline Word() {} // used in placement new
  inline void* operator new(size_t, WordList* garbage) {
    Word* p = garbage->get();
    return p ? p : ::new Word;
  }

  // calculate and store width of word
  inline void set_width(int* char_width) {
    const char* w = m_word - 1;
    m_width = 0;
    while (*++w != 0) m_width += char_width[size_t(*w)];
  }

  inline const char* word()  const { return m_word; }
  inline int         size()  const { return m_size; }
  inline int         width() const { return m_width; }
  inline void        width(int w)  { m_width = w; }
  inline void        accu(int s)   { m_accu = s; }
  inline int         accu()       const { return m_accu; }
  inline bool        syllable()   const { return m_syllable; }

  // find position after the word, if at end of line:
  inline int break_pos(int hyp_width) const { // position of break after word
    return m_accu + m_width + (m_syllable ? hyp_width : 0);
  }

  // cut due to hyphenation
  inline void cut(char* s, int* char_width) {
    // cut at s, preceeding becomes syllable
    *s = 0;
    m_size = s - m_word;
    m_syllable = true;
    set_width(char_width);
  }

  inline void print_test() { printf("%s%c\n",m_word,m_syllable ? '-' : ' '); }
  inline void print(bool end) {
    if (end) {
      if (m_syllable) printf("%s-\n",m_word);
      else            printf("%s\n",m_word);
      // while (gap--) printf(" ");
      // printf("|\n");
    }
    else     {
      if (m_syllable) printf("%s",m_word);
      else            printf("%s ",m_word);
    }
  }
}; // Word

class Break : public slnk {
  Break* m_pred;
  Word*  m_word;  // word after which we break
  int    m_cost;  // accumulated cost along shortest path
  int    m_gap;   // maximum gap along shorest path
  int    m_fanin;
  int    m_nr;    // for debugging
public:
  inline Break(Break* pred, Word* word, int cost, int gap) :
         m_pred(pred),
         m_word(word),
         m_cost(cost),
         m_gap(gap),
         m_fanin(0) {
    static size_t nr = 0;
    m_nr = nr++;
  }
  inline Break() {} // used in placement new
  inline void* operator new(size_t, BreakList* garbage) {
    Break* p = garbage->get();
    return p ? p : ::new Break;
  }
  inline Break*pred()   const { return m_pred; }
  inline void  pred(Break* b) { m_pred = b; }
  inline Word* word()   const { return m_word; }
  inline int   cost()   const { return m_cost; }
  inline void  cost(int c)    { m_cost = c; }
  inline int   gap()    const { return m_gap; }
  inline void  gap(int c)     { m_gap = c; }
  inline int   fanin()  const { return m_fanin; }
  inline void  incr_fanin()   { m_fanin++; }
  inline void  print(const char* msg) const {
    int pre_nr = m_pred ? m_pred->m_nr : -1;
    printf("%s break %d --> %d cost %3d path_gap %d after ",
           msg,m_nr,pre_nr,m_cost,m_gap); 
    m_word->print_test();
  }
}; // Break

class Formatter {
  FILE*                   m_in;
  char                    m_line[256];       // buffer for one input line
  int                     m_char_width[256]; // width of chars
  int                     m_space_width;     // width of space
  int                     m_hyphen_width;    // width of hyphen
  WordList   m_words;     // list of words or syllables of paragraph
  WordList   m_word_garbage; 
  BreakList  m_breaks;    // all breaks
  BreakList  m_break_garbage; 
  inline int get_line() {
    char* s = fgets(m_line,sizeof(m_line),m_in);
    return s ? strlen(s) : 0;
  }
  inline bool vowel(char c) {
    static const char* vowels = "aeiouäöü";
    return  strchr(vowels,c) != 0;
  }
  bool hyphenate(Word* w, int limit);
  int  greedy(int line_length);

  inline int break_position(Break* b) const {
    Word* w = m_words.succ(b->word());
    return w->accu();
  }

  Break* solve_kp(int line_length, int upper_bound);
  void output(Break* best, int line_length);
public:
  inline Formatter() : m_in(0),
                       m_space_width(0),
                       m_hyphen_width(0) {
  }
  inline ~Formatter() { // free allocated memory
    m_word_garbage.destroy();
    m_break_garbage.destroy();
  }

  void store_font(FILE* font);
  void store_paragraph(FILE* paragraph);

  inline void print_ruler(int length) {
    length /= m_space_width;
    while (length--) printf("*"); printf("\n");
  }

  inline void run(int line_length) {

    line_length *= m_space_width;

    timer t;

    int max_gap = greedy(line_length);
    Break* best = solve_kp(line_length,max_gap);
    output(best,line_length);

    // put Word and Break objects int garbage for reuse:
    m_word_garbage.conc(m_words);   
    m_break_garbage.conc(m_breaks);
    
    t.used_time("cpu time for solution");
  }
}; // Formatter

bool Formatter::hyphenate(Word* w, int limit) {
  const char* start = w->word();
  const char* stop  = start + w->size();
  // find first vowel after start
  const char* s = start + 1;
  while (s != stop) {
    while (s != stop && ! vowel(*s)) s++;
    // now *s is a vowel and *(s-1) a consonant
    if (s - start >= limit && stop - s >= 2) { // accept
      // printf("----- word: %s s: %s\n",start,s); 
      m_words.append(new (&m_word_garbage) Word(s-1,m_char_width,false));
      w->cut((char*)s - 1, m_char_width);
      return true;
    }
    else { // skip following vowels
      while (s != stop && vowel(*s)) s++;
    }
  }
  return false;
} // hyphenate


// store widths of all simple chars:
void Formatter::store_font(FILE* font) {
  m_in = font;
  int length;
  // skip till "spacewidth"
  while ((length = get_line()) != 0) {
    if (!strncmp(m_line,"spacewidth",strlen("spacewidth"))) {
      m_space_width = m_char_width[' '] = atoi(m_line + strlen("spacewidth "));
      printf("space width %d\n",m_space_width);
      break;
    }
  }

  // skip till "charset"
  while ((length = get_line()) != 0) {
    if (strncmp(m_line,"charset",strlen("charset")) == 0) break;
  }

  while ((length = get_line()) != 0) {
    if (m_line[1] == '\t') { // we store only simple chars
      size_t charval = size_t(m_line[0]);
      m_char_width[charval] = atoi(m_line + 2);
      // printf("char %c width %d\n",m_line[0], m_char_width[charval]);
    }
  }
  m_hyphen_width = m_char_width[size_t('-')];
} //store_font

void Formatter::store_paragraph(FILE* paragraph) {
/* the whole input file is considered one paragraph, so we read lines
   till EOF is seen
*/
  m_in = paragraph;
  int length;
  while ((length = get_line()) != 0) {
    char* s    = m_line;
    char* stop = s + length;
    assertion(*(stop-1) == '\n',"newline expected");
    *(stop-1) = ' ';
    while (s < stop) {
      // skip blanks: FIXME?
      while (s < stop && *s == ' ') s++;
      // find end of word:
      char* e = s + 1;
      while (e < stop && *e != ' ') e++;
      // now e is behind word
      *e = 0;
      m_words.append(new(&m_word_garbage) Word(s,m_char_width,false));
      s = e + 1;
      Word* w = m_words.last();
      int limit = 3;
      if (w->size() >= 5) {
        while (hyphenate(w,limit)) {
          // w->print();
          w = m_words.last();
          limit = 2;
        }
        // if (limit != 3) w->print();
      }
    }
  }
  // for (Word* tp = m_words.first(); tp; tp = m_words.succ(tp)) tp->print();
} // store_paragraph

int Formatter::greedy(int limit) {
/*------------------------------------
  we try to fill line by line as much as possible and report
  the maximum gap, to be used as upper bound im solve_kp
*/ 
  timer t;
  int length  = 0;
  printf("greedy result:\n");
  // print_ruler(limit); // for debugging
  int lines   = 0;
  int max_gap = 0;
  Word*  pred = 0;
  for (Word* w = m_words.first(); w; w = m_words.succ(pred = w)) { // all words
    int old_length = length;
    length += w->width();
    int length_with_hyphen = length + (w->syllable() ? m_hyphen_width : 0);
    // printf("%s length %d limit %d\n",w->word(),length_with_hyphen,limit);
    if (length_with_hyphen > limit) { // to long
      // old_length contains trailing blank, unless syllable
      if (pred->syllable()) old_length += m_hyphen_width; // need '-'
      else                  old_length -= m_space_width;  // don't need blank
      int gap = limit - old_length;
      maximize(max_gap,gap);
      length = w->width();
      pred->print(true);
      printf(".br\n");
      lines++;
    }
    else { // word fits in
      if (pred) pred->print(false);
    }
    if (! w->syllable()) length += m_space_width; // add trailing blank 
  }
  m_words.last()->print(false);
  printf("\nmax_gap: %d lines %d\n",max_gap,lines+1);
  t.used_time("  cpu time for greedy");
  return max_gap;
} // greedy

Break* Formatter::solve_kp(int limit, int upper_bound) {
/* a dynamic program along the lines of Knuth / Pass, but greatly
 * simplified
*/
  timer t;

  Word* first_word = m_words.first();
  Word* last_word  = m_words.last();

  Word dummy_word("dummy",0,false); 
  m_words.push(&dummy_word);

  // dummy_break will be last in optimum chain of Breaks:
  Break* dummy_break = new(&m_break_garbage) Break(0,&dummy_word,0,0);
  m_breaks.push(dummy_break);

  int total = 0;
  for (Word* w = first_word; w; w = m_words.succ(w)) { // all words
    
    // for each word, we try to generate a Break

    w->accu(total); // total length of words seen including trailing blanks

    total += w->width();
    if (! w->syllable()) total += m_space_width; // add blank

    // printf("----------------------------------------\n");
    // w->print_test();
    // printf("total = %3d %s %d\n",w->accu(),w->word(),w->width());


    // find legal preceding breaks, and select best one:
    Break* best_break = 0;
    int    best_gap   = INT_MAX;
    int    best_cost  = INT_MAX;
    for (Break* b = m_breaks.first(); b; b = m_breaks.succ(b)) {
      // b->print("   trying ");
      int length = w->break_pos(m_hyphen_width) - break_position(b);
      int gap = limit - length;
      // line too short?
      if (w != last_word && gap > upper_bound) continue;
      // printf("gap %d\n",gap);
      // line too long?
      if (length > limit) {
        assertion(b != m_breaks.first(),"no break possible");
        // breaks without fanin could be eliminated,
        // but won't be used anyway
        if (b->fanin() == 0) printf("no fanin\n");
        break;
      }
      if (w == last_word) gap = 0; // for last line gap is irrelevant
      
      int path_gap  = max(b->gap(),gap);   // maximum gap along path
      int path_cost = b->cost() + gap*gap; // total cost along path
      if (minimize(best_gap,path_gap)) {
        best_cost = path_cost;
        best_break = b;
      }
      else {
        if (path_gap == best_gap) {
          if (minimize(best_cost,path_cost)) {
            best_break = b;
          }
        }
      }
      // printf("   length %d\n",length);
    } // all feasible breaks

    if (best_break) { // we keep only one break per word!
      // best_break->print("   using ");
      m_breaks.push(new(&m_break_garbage)
          Break(best_break,w,best_cost,best_gap));
      best_break->incr_fanin();
      // m_breaks.first()->print("creating ");
    }

  } // all words

  m_words.get(); // remove dummy

  printf("total breaks: %d\n",m_breaks.size());
  // now find best break, s.t. word == last_word:
  Break* best_break  = m_breaks.first();
  printf("breaks/words %.1f\n",double(m_breaks.size())/m_words.size());


#if 0
  Break* b = best_break;
  while (b) {
    b->print("in chain ");
    b = b->pred();
  }
#endif

  t.used_time("  cpu time for KP");

  // for testing, print line_length *'s:
  printf("dynamic programming solution:\n");
  // print_ruler(limit);


  return best_break;

} // solve_kp

void Formatter::output(Break* best, int line_length) {

  printf("best->gap %d\n",best->gap());

  // first we reverse the pred-chain of Breaks:
  Break* tp = best;
  Break* pred = tp->pred();
  Break* next = 0;
  while ((next = pred->pred()) != 0) {
    pred->pred(tp);
    tp = pred;
    pred = next;
  }

  // now tp is first in chain
  int length  = 0;
  int max_gap = 0;
  int lines   = 0;
  Word*  last_word = m_words.last();
  Break* b = tp;
  for (Word* w = m_words.first(); w; w = m_words.succ(w)) {
    length += w->width();
    if (w == b->word()) { // end of line
      if (w->syllable()) length += m_hyphen_width;
      int gap = line_length - length;
      if (w != last_word) maximize(max_gap,gap);
      w->print(true);
      printf(".br\n");
      b = b->pred();
      length = 0;
      lines++;
    }
    else {
      if (! w->syllable()) length += m_space_width;
      w->print(false);
    }
  }
  printf("\nmax_gap: %d lines: %d\n",max_gap,lines);
} // output

int main (int argc, char** argv) {

  assertion(argc == 3,"need file name and line_length");

  const char* file_name = argv[1];
  int line_length = 0;
  sscanf(argv[2],"%d",&line_length);
  printf("line_length %d\n",line_length);

  FILE* paragraph = fopen(file_name,"r");
  FILE* font      = fopen("HR","r");

  Formatter my_formatter;
  my_formatter.store_font(font);           // store width of chars
  my_formatter.store_paragraph(paragraph); // make list of words/syllables

  my_formatter.run(line_length);

  return 0;
}
