#ifndef _ILIST_H
#define _ILIST_H

#include "assertion.h"

/********************************************
  This is an implementation of linked lists of objects that are 
  derived from slnk or dlnk for singly or doubly linked lists, repectively.
  So, the links become part of the objects, no auxiliary objects are
  allocated / deallocated when an object is added to / removed from a list.
*/

class slnk { // link class for singly linked lists
  inline slnk(const slnk&) {};
  inline slnk& operator=(const slnk&) { return *this; }
  slnk* m_next;
public:
  inline slnk()         : m_next(0) {}
  inline slnk(slnk* p)  : m_next(p) {}
  inline ~slnk() {
    assertion(! m_next,"attempt to delete active slist element");
  }
  inline bool  linked() const { return m_next != 0; }
  inline slnk* next()         { return m_next; }
  inline slnk* next(slnk* s)  { return m_next = s; }
  inline slnk* last()         {
    assertion(false,"illegal access to last() in singly linked list");
    return 0;
  }
  inline slnk* last(slnk*) {
    assertion(false,"illegal access to last() in singly linked list");
    return 0;
  }
}; // slnk

class dlnk { // link class for doubly linked lists
  inline dlnk(const dlnk&) {};
  inline dlnk& operator=(const dlnk&) { return *this; }
  dlnk* m_next;
  dlnk* m_last;
public:
  inline dlnk()         : m_next(0) {}
  inline dlnk(dlnk* p)  : m_next(p) {}
  inline ~dlnk() {
    assertion(! m_next,"attempt to delete active slist element");
  }
  inline bool  linked() const { return m_next != 0; }
  inline dlnk* next()         { return m_next; }
  inline dlnk* next(dlnk* s)  { return m_next = s; }
  inline dlnk* last()         { return m_last; }
  inline dlnk* last(dlnk* s)  { return m_last = s; }
}; // dlnk

/****************************************
  T is the type of objects the list contains
  L is the link class used
  D is true / false for singly / dobly linked lists
*/
template <class T, class L, bool D>
class ilist {
  inline ilist(const ilist<T,L,D>&) {
  }
  inline int operator=(const ilist<T,L,D>&) {
    return 0;
  }
  size_t m_elements; // number of elements in the list
  L*     m_tail;     // last element, points to first

  inline L* priv_push(L* a) {
    assertion(! a->next(),"a already in list!");
    if (m_tail) {
      if (D) { // doubly linked
        a->next(m_tail->next());
        a->last(m_tail);
        a->next()->last(m_tail->next(a));
      }
      else { // singly linked
        a->next(m_tail->next());
        m_tail->next(a);
      }
    }
    else { // was empty
      a->next(m_tail = a);
      if (D) a->last(a);
    }
    m_elements++;
    return a;
  }
  inline L* priv_append(L* a) {
    priv_push(a);
    return (m_tail = a);
  }

  inline L* priv_fast_pop() { // we know that m_tail != 0
    L* f = m_tail->next();
    if (D) { // dlist
      if (f == m_tail) m_tail = 0;
      else {
        m_tail->next(f->next());
        f->next()->last(m_tail);
      }
      f->last(0);
    }
    else { // slist
      if (f == m_tail) m_tail = 0; else m_tail->next(f->next());
    }
    f->next(0); // mark not on list
    return f;
  }

  inline L* priv_first() const { return m_tail ? m_tail->next() : 0; }

  inline L* priv_last()  const { return m_tail; }

  inline L* priv_succ(L* a) const {
    return (a == m_tail) ? 0 : a->next();
  }

  inline L* priv_pred(const L* a) {
    if (D) { // dlist
      return (a == m_tail->next()) ? 0 : a->last();
    }
    else {
      assertion(false,"pred not supported for slist");
      return 0;
    }
  }
public:
  inline ilist() : m_elements(0), m_tail(0) { // constructor for empty list
  }
  inline ~ilist() { // destructor
    clear();
  }
  // add element a as first element
  inline T* push(const T* a) {
    return (T*) priv_push((L*)a);
  }
  
  // add element a as last element
  inline T* append(const T* a) {
    return (T*) priv_append((L*)a);
  }
  
  // remove and return first element
  inline T* fast_pop() { // we know that m_tail != 0
    return (T*) priv_fast_pop();
  }

  // remove and return first element, if available; else return 0
  inline T* get() {
    return m_tail ? fast_pop() : 0;
  } 

  // return first element, or zero if list is empty
  inline T* first()          const { return (T*) priv_first(); }

  // return last element, or zero if list is empty
  inline T* last()           const { return (T*) priv_last (); }

  // return successor of a, or zero if a is the last element
  inline T* succ(const T* a) const { return (T*) priv_succ ((L*) a); }

  // return predecessor of a, or zero if a is the first element
  inline T* pred(const T* a) const { return (T*) priv_pred ((L*) a); }

  // concatenate list l1 to this list, l1 becomes empty
  inline void conc(ilist<T,L,D>& l1) {
    if (! l1.m_tail) return;

    m_elements += l1.m_elements;
    l1.m_elements = 0;

    if (! m_tail) {
      m_tail     = l1.m_tail;
      l1.m_tail  = 0;
      return;
    }
    if (D) { // dlist
      L* head1      = m_tail->next();
      L* head2      = l1.m_tail->next();
      m_tail->next(head2);
      head2->last(m_tail);
      l1.m_tail->next(head1);
      m_tail = head1->last(l1.m_tail);
    }
    else { // slist
      L* tmp        =  l1.m_tail->next();
      l1.m_tail->next(m_tail->next());
      m_tail->next(tmp);
      m_tail = l1.m_tail;
    }
    l1.m_tail = 0;
  } // conc

  // revese the sequence of elements in the list
  inline void reverse() {
    if (m_tail == 0 || m_tail->next() == m_tail) return;
    L* last = m_tail;
    L* head = m_tail->next();
    L* next = 0;
    for (L* tp = head; next != head; tp = next) {
      next = tp->next();
      if (D) { // dlist
        tp->last(next);
      }
      tp->next(last);
      last = tp;
    }
    m_tail = head;
  } // reverse

  // make list empty, keep elements
  inline void clear() {
    L* p1 = m_tail;
    while (p1) {
      L* p2 = p1->next(); 
      p1->next(0);
      if (D) p1->last(0);
      p1 = p2;
    }
    m_tail     = 0;
    m_elements = 0;
  } // clear

  // make list empty, delete elements
  inline void destroy() {
    T* tp;
    while ((tp = get())) delete tp;
  } // destroy

  // return number of elements currently in the list
  size_t size() { return m_elements; }

};

#endif
